"""This module contains the definition of validation errors and warnings raised during graph methods"""

from cognite.neat.constants import DEFAULT_DOCS_URL
from cognite.neat.exceptions import NeatException

DOCS_BASE_URL = f"{DEFAULT_DOCS_URL}api/exceptions.html#{__name__}"


class UnsupportedPropertyType(NeatException):
    """Unsupported property type when processing the graph capturing sheet

    Args:
        property_type: property type that is not supported
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_: str = "UnsupportedPropertyType"
    code: int = 1000
    description: str = "Unsupported property type when processing the graph capturing sheet."
    example: str = ""
    fix: str = ""

    def __init__(self, property_type: str, verbose: bool = False):
        self.property_type = property_type

        self.message = (
            f"Property type {self.property_type} is not supported. "
            " Only the following property types are supported: DatatypeProperty and ObjectProperty"
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class NamespaceRequired(NeatException):
    """The functionality requires namespace in the TransformationRules.

    Args:
        functionality: functionality that requires namespace
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_ = "NamespaceRequired"
    description: str = "The functionality requires namespace in the TransformationRules."
    example: str = ""
    fix: str = ""

    def __init__(self, functionality: str, verbose: bool = False):
        self.message = (
            f"Namespace is required to be set in the Transformation rules"
            f"to use {functionality}."
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)


class DatasetIdRequired(NeatException):
    """The functionality requires data_set_id in the TransformationRules.

    Args:
        functionality: functionality that requires namespace
        verbose: flag that indicates whether to provide enhanced exception message, by default False
    """

    type_ = "DatasetIdRequired"
    description: str = "The functionality requires data_set_id in the TransformationRules."
    example: str = ""
    fix: str = ""

    def __init__(self, functionality: str, verbose: bool = False):
        self.message = (
            f"DataSetId is required to be set in the Transformation rules"
            f"to use {functionality}."
            f"\nFor more information visit: {DOCS_BASE_URL}.{self.__class__.__name__}"
        )

        if verbose:
            self.message += f"\nDescription: {self.description}"
            self.message += f"\nExample: {self.example}"
            self.message += f"\nFix: {self.fix}"
        super().__init__(self.message)



from pathlib import Path
from typing import Literal

import yaml

from ._dict2rules import ArbitraryDictImporter


class ArbitraryYAMLImporter(ArbitraryDictImporter):
    """
    Importer for data given in a YAML file or string.

    This importer infers the data model from the YAML string based on the shape of the data.

    Args:
        yaml_path_or_str: Path to file with YAML.
        relationship_direction: Direction of relationships, either "parent-to-child" or "child-to-parent". JSON
            files are nested with children nested inside parents. This option determines whether the resulting rules
            will have an edge from parents to children or from children to parents.
    """

    def __init__(
        self,
        yaml_path_or_str: Path,
        relationship_direction: Literal["parent-to-child", "child-to-parent"] = "parent-to-child",
    ):
        if isinstance(yaml_path_or_str, str):
            data = yaml.safe_load(yaml_path_or_str)
            super().__init__(data, relationship_direction)
        elif isinstance(yaml_path_or_str, Path):
            if not yaml_path_or_str.exists():
                raise ValueError(f"File {yaml_path_or_str} does not exist")
            if yaml_path_or_str.suffix != ".json":
                raise ValueError(f"File {yaml_path_or_str} is not a JSON file")
            self.json_path = yaml_path_or_str
            data = yaml.safe_load(yaml_path_or_str.read_text())
            super().__init__(data, relationship_direction)
        else:
            raise TypeError(f"Expected Path or str, got {type(yaml_path_or_str)}")



from ._base import BaseExtractor
from ._dexpi import DexpiXML
from ._graph_capturing_sheet import GraphCapturingSheet
from ._mock_graph_generator import MockGraphGenerator

__all__ = ["BaseExtractor", "MockGraphGenerator", "GraphCapturingSheet", "DexpiXML"]



from collections.abc import Callable, Sequence
from graphlib import TopologicalSorter
from typing import Any, Literal, cast

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    T_CogniteResourceList,
    T_WritableCogniteResource,
    T_WriteClass,
)
from cognite.client.data_classes.data_modeling import (
    Container,
    ContainerApply,
    ContainerApplyList,
    ContainerList,
    DataModel,
    DataModelApply,
    DataModelApplyList,
    DataModelList,
    RequiresConstraint,
    Space,
    SpaceApply,
    SpaceApplyList,
    SpaceList,
    View,
    ViewApply,
    ViewApplyList,
    ViewList,
)
from cognite.client.data_classes.data_modeling.ids import (
    ContainerId,
    DataModelId,
    ViewId,
)
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr

from ._base import T_ID, ResourceLoader, T_WritableCogniteResourceList


class DataModelingLoader(
    ResourceLoader[T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList]
):
    def in_space(self, item: T_WriteClass | T_WritableCogniteResource, space: set[str]) -> bool:
        if hasattr(item, "space"):
            return item.space in space
        raise ValueError(f"Item {item} does not have a space attribute")

    def sort_by_dependencies(self, items: list[T_WriteClass]) -> list[T_WriteClass]:
        return items

    def _create_force(
        self,
        items: Sequence[T_WriteClass],
        tried_force_deploy: set[T_ID],
        create_method: Callable[[Sequence[T_WriteClass]], T_WritableCogniteResourceList],
    ) -> T_WritableCogniteResourceList:
        try:
            return create_method(items)
        except CogniteAPIError as e:
            failed_items = {failed.as_id() for failed in e.failed if hasattr(failed, "as_id")}
            to_redeploy = [
                item
                for item in items
                if item.as_id() in failed_items and item.as_id() not in tried_force_deploy  # type: ignore[attr-defined]
            ]
            if not to_redeploy:
                # Avoid infinite loop
                raise e
            ids = [item.as_id() for item in to_redeploy]  # type: ignore[attr-defined]
            tried_force_deploy.update(ids)
            self.delete(ids)
            return self._create_force(to_redeploy, tried_force_deploy, create_method)


class SpaceLoader(DataModelingLoader[str, SpaceApply, Space, SpaceApplyList, SpaceList]):
    resource_name = "spaces"

    @classmethod
    def get_id(cls, item: Space | SpaceApply) -> str:
        return item.space

    def create(self, items: Sequence[SpaceApply]) -> SpaceList:
        return self.client.data_modeling.spaces.apply(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> SpaceList:
        return self.client.data_modeling.spaces.retrieve(ids)

    def update(self, items: Sequence[SpaceApply]) -> SpaceList:
        return self.create(items)

    def delete(self, ids: SequenceNotStr[str] | Sequence[Space | SpaceApply]) -> list[str]:
        if all(isinstance(item, Space) for item in ids) or all(isinstance(item, SpaceApply) for item in ids):
            ids = [cast(Space | SpaceApply, item).space for item in ids]
        return self.client.data_modeling.spaces.delete(cast(SequenceNotStr[str], ids))


class ViewLoader(DataModelingLoader[ViewId, ViewApply, View, ViewApplyList, ViewList]):
    resource_name = "views"

    def __init__(self, client: CogniteClient, existing_handling: Literal["fail", "skip", "update", "force"] = "fail"):
        super().__init__(client)
        self.existing_handling = existing_handling
        self._cache_view_by_id: dict[ViewId, View] = {}
        self._tried_force_deploy: set[ViewId] = set()

    @classmethod
    def get_id(cls, item: View | ViewApply) -> ViewId:
        return item.as_id()

    def create(self, items: Sequence[ViewApply]) -> ViewList:
        if self.existing_handling == "force":
            return self._create_force(items, self._tried_force_deploy, self.client.data_modeling.views.apply)
        else:
            return self.client.data_modeling.views.apply(items)

    def retrieve(self, ids: SequenceNotStr[ViewId]) -> ViewList:
        return self.client.data_modeling.views.retrieve(cast(Sequence, ids))

    def update(self, items: Sequence[ViewApply]) -> ViewList:
        return self.create(items)

    def delete(self, ids: SequenceNotStr[ViewId]) -> list[ViewId]:
        return self.client.data_modeling.views.delete(cast(Sequence, ids))

    def _as_write_raw(self, view: View) -> dict[str, Any]:
        dumped = view.as_write().dump()
        if view.properties:
            # All read version of views have all the properties of their parent views.
            # We need to remove these properties to compare with the local view.
            parents = self._retrieve_view_ancestors(view.implements or [], self._cache_view_by_id)
            for parent in parents:
                for prop_name in parent.properties.keys():
                    dumped["properties"].pop(prop_name, None)

        if "properties" in dumped and not dumped["properties"]:
            # All properties were removed, so we remove the properties key.
            dumped.pop("properties", None)
        return dumped

    def are_equal(self, local: ViewApply, remote: View) -> bool:
        local_dumped = local.dump()
        if not remote.implements:
            return local_dumped == remote.as_write().dump()

        cdf_resource_dumped = self._as_write_raw(remote)

        if "properties" in local_dumped and not local_dumped["properties"]:
            # In case the local properties are set to an empty dict.
            local_dumped.pop("properties", None)

        return local_dumped == cdf_resource_dumped

    def as_write(self, view: View) -> ViewApply:
        return ViewApply.load(self._as_write_raw(view))

    def retrieve_all_parents(self, views: list[ViewId]) -> list[View]:
        return self._retrieve_view_ancestors(views, self._cache_view_by_id)

    def _retrieve_view_ancestors(self, parents: list[ViewId], cache: dict[ViewId, View]) -> list[View]:
        """Retrieves all ancestors of a view.

        This will mutate the cache passed in, and return a list of views that are the ancestors
        of the views in the parents list.

        Args:
            parents: The parents of the view to retrieve all ancestors for
            cache: The cache to store the views in
        """
        parent_ids = parents.copy()
        found: list[View] = []
        found_ids: set[ViewId] = set()
        while parent_ids:
            to_lookup: set[ViewId] = set()
            grand_parent_ids = []
            for parent in parent_ids:
                if parent in found_ids:
                    continue
                elif parent in cache:
                    found.append(cache[parent])
                    grand_parent_ids.extend(cache[parent].implements or [])
                else:
                    to_lookup.add(parent)

            if to_lookup:
                looked_up = self.client.data_modeling.views.retrieve(list(to_lookup))
                cache.update({view.as_id(): view for view in looked_up})
                found.extend(looked_up)
                found_ids.update({view.as_id() for view in looked_up})
                for view in looked_up:
                    grand_parent_ids.extend(view.implements or [])

            parent_ids = grand_parent_ids
        return found


class ContainerLoader(DataModelingLoader[ContainerId, ContainerApply, Container, ContainerApplyList, ContainerList]):
    resource_name = "containers"

    def __init__(self, client: CogniteClient, existing_handling: Literal["fail", "skip", "update", "force"] = "fail"):
        super().__init__(client)
        self.existing_handling = existing_handling
        self._tried_force_deploy: set[ContainerId] = set()

    @classmethod
    def get_id(cls, item: Container | ContainerApply) -> ContainerId:
        return item.as_id()

    def sort_by_dependencies(self, items: Sequence[ContainerApply]) -> list[ContainerApply]:
        container_by_id = {container.as_id(): container for container in items}
        container_dependencies = {
            container.as_id(): {
                const.require
                for const in container.constraints.values()
                if isinstance(const, RequiresConstraint) and const.require in container_by_id
            }
            for container in items
        }
        return [
            container_by_id[container_id] for container_id in TopologicalSorter(container_dependencies).static_order()
        ]

    def create(self, items: Sequence[ContainerApply]) -> ContainerList:
        if self.existing_handling == "force":
            return self._create_force(items, self._tried_force_deploy, self.client.data_modeling.containers.apply)
        else:
            return self.client.data_modeling.containers.apply(items)

    def retrieve(self, ids: SequenceNotStr[ContainerId]) -> ContainerList:
        return self.client.data_modeling.containers.retrieve(cast(Sequence, ids))

    def update(self, items: Sequence[ContainerApply]) -> ContainerList:
        return self.create(items)

    def delete(self, ids: SequenceNotStr[ContainerId]) -> list[ContainerId]:
        return self.client.data_modeling.containers.delete(cast(Sequence, ids))

    def are_equal(self, local: ContainerApply, remote: Container) -> bool:
        local_dumped = local.dump(camel_case=True)
        if "usedFor" not in local_dumped:
            # Setting used_for to "node" as it is the default value in the CDF.
            local_dumped["usedFor"] = "node"

        return local_dumped == remote.as_write().dump(camel_case=True)


class DataModelLoader(DataModelingLoader[DataModelId, DataModelApply, DataModel, DataModelApplyList, DataModelList]):
    resource_name = "data_models"

    @classmethod
    def get_id(cls, item: DataModel | DataModelApply) -> DataModelId:
        return item.as_id()

    def create(self, items: Sequence[DataModelApply]) -> DataModelList:
        return self.client.data_modeling.data_models.apply(items)

    def retrieve(self, ids: SequenceNotStr[DataModelId]) -> DataModelList:
        return self.client.data_modeling.data_models.retrieve(cast(Sequence, ids))

    def update(self, items: Sequence[DataModelApply]) -> DataModelList:
        return self.create(items)

    def delete(self, ids: SequenceNotStr[DataModelId]) -> list[DataModelId]:
        return self.client.data_modeling.data_models.delete(cast(Sequence, ids))

    def are_equal(self, local: DataModelApply, remote: DataModel) -> bool:
        local_dumped = local.dump()
        cdf_resource_dumped = remote.as_write().dump()

        # Data models that have the same views, but in different order, are considered equal.
        # We also account for whether views are given as IDs or View objects.
        local_dumped["views"] = sorted(
            (v if isinstance(v, ViewId) else v.as_id()).as_tuple() for v in local.views or []
        )
        cdf_resource_dumped["views"] = sorted(
            (v if isinstance(v, ViewId) else v.as_id()).as_tuple() for v in remote.views or []
        )

        return local_dumped == cdf_resource_dumped



from fastapi import APIRouter
from prometheus_client import REGISTRY

router = APIRouter()


@router.get("/api/metrics")
def get_metrics():
    metrics = REGISTRY.collect()
    return {"prom_metrics": metrics}



from ._field import (
    ExternalIdType,
    NamespaceType,
    PrefixType,
    PropertyType,
    StrListType,
    StrOrListType,
    VersionType,
)

__all__ = [
    "StrOrListType",
    "StrListType",
    "NamespaceType",
    "PrefixType",
    "ExternalIdType",
    "VersionType",
    "PropertyType",
]



from dataclasses import dataclass


@dataclass
class UnitEntry:
    semantic_type: str
    unit_type: str
    unit: str


_UNIT_ENTRIES = [
    UnitEntry("Acceleration", "AccelerationUnit", "centimetrePerSecondSquared"),
    UnitEntry("Acceleration", "AccelerationUnit", "gForce"),
    UnitEntry("Acceleration", "AccelerationUnit", "metrePerSecondSquared"),
    UnitEntry("Angle", "AngleUnit", "degreeOfArc"),
    UnitEntry("Angle", "AngleUnit", "minuteOfArc"),
    UnitEntry("Angle", "AngleUnit", "radian"),
    UnitEntry("Angle", "AngleUnit", "secondOfArc"),
    UnitEntry("Angle", "AngleUnit", "turn"),
    UnitEntry("AngularAcceleration", "AngularAccelerationUnit", "radianPerSecondSquared"),
    UnitEntry("AngularVelocity", "AngularVelocityUnit", "degreePerSecond"),
    UnitEntry("AngularVelocity", "AngularVelocityUnit", "radianPerSecond"),
    UnitEntry("AngularVelocity", "AngularVelocityUnit", "revolutionPerMinute"),
    UnitEntry("AngularVelocity", "AngularVelocityUnit", "revolutionPerSecond"),
    UnitEntry("Area", "AreaUnit", "acre"),
    UnitEntry("Area", "AreaUnit", "hectare"),
    UnitEntry("Area", "AreaUnit", "squareCentimetre"),
    UnitEntry("Area", "AreaUnit", "squareFoot"),
    UnitEntry("Area", "AreaUnit", "squareInch"),
    UnitEntry("Area", "AreaUnit", "squareKilometre"),
    UnitEntry("Area", "AreaUnit", "squareMetre"),
    UnitEntry("Area", "AreaUnit", "squareMillimetre"),
    UnitEntry("Capacitance", "CapacitanceUnit", "farad"),
    UnitEntry("Capacitance", "CapacitanceUnit", "microfarad"),
    UnitEntry("Capacitance", "CapacitanceUnit", "millifarad"),
    UnitEntry("Capacitance", "CapacitanceUnit", "nanofarad"),
    UnitEntry("Capacitance", "CapacitanceUnit", "picofarad"),
    UnitEntry("Current", "CurrentUnit", "ampere"),
    UnitEntry("Current", "CurrentUnit", "microampere"),
    UnitEntry("Current", "CurrentUnit", "milliampere"),
    UnitEntry("DataRate", "DataRateUnit", "bitPerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "bytePerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "exbibitPerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "exbibytePerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "gibibitPerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "gibibytePerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "kibibitPerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "kibibytePerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "mebibitPerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "mebibytePerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "tebibitPerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "tebibytePerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "yobibitPerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "yobibytePerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "zebibitPerSecond"),
    UnitEntry("DataRate", "DataRateUnit", "zebibytePerSecond"),
    UnitEntry("DataSize", "DataSizeUnit", "bit"),
    UnitEntry("DataSize", "DataSizeUnit", "byte"),
    UnitEntry("DataSize", "DataSizeUnit", "exbibit"),
    UnitEntry("DataSize", "DataSizeUnit", "exbibyte"),
    UnitEntry("DataSize", "DataSizeUnit", "gibibit"),
    UnitEntry("DataSize", "DataSizeUnit", "gibibyte"),
    UnitEntry("DataSize", "DataSizeUnit", "kibibit"),
    UnitEntry("DataSize", "DataSizeUnit", "kibibyte"),
    UnitEntry("DataSize", "DataSizeUnit", "mebibit"),
    UnitEntry("DataSize", "DataSizeUnit", "mebibyte"),
    UnitEntry("DataSize", "DataSizeUnit", "tebibit"),
    UnitEntry("DataSize", "DataSizeUnit", "tebibyte"),
    UnitEntry("DataSize", "DataSizeUnit", "yobibit"),
    UnitEntry("DataSize", "DataSizeUnit", "yobibyte"),
    UnitEntry("DataSize", "DataSizeUnit", "zebibit"),
    UnitEntry("DataSize", "DataSizeUnit", "zebibyte"),
    UnitEntry("Density", "DensityUnit", "gramPerCubicMetre"),
    UnitEntry("Density", "DensityUnit", "kilogramPerCubicMetre"),
    UnitEntry("Distance", "LengthUnit", "astronomicalUnit"),
    UnitEntry("Distance", "LengthUnit", "centimetre"),
    UnitEntry("Distance", "LengthUnit", "foot"),
    UnitEntry("Distance", "LengthUnit", "inch"),
    UnitEntry("Distance", "LengthUnit", "kilometre"),
    UnitEntry("Distance", "LengthUnit", "metre"),
    UnitEntry("Distance", "LengthUnit", "micrometre"),
    UnitEntry("Distance", "LengthUnit", "mile"),
    UnitEntry("Distance", "LengthUnit", "millimetre"),
    UnitEntry("Distance", "LengthUnit", "nanometre"),
    UnitEntry("Distance", "LengthUnit", "nauticalMile"),
    UnitEntry("ElectricCharge", "ChargeUnit", "coulomb"),
    UnitEntry("Energy", "EnergyUnit", "electronvolt"),
    UnitEntry("Energy", "EnergyUnit", "gigajoule"),
    UnitEntry("Energy", "EnergyUnit", "joule"),
    UnitEntry("Energy", "EnergyUnit", "kilojoule"),
    UnitEntry("Energy", "EnergyUnit", "kilowattHour"),
    UnitEntry("Energy", "EnergyUnit", "megaelectronvolt"),
    UnitEntry("Energy", "EnergyUnit", "megajoule"),
    UnitEntry("Force", "ForceUnit", "newton"),
    UnitEntry("Force", "ForceUnit", "ounce"),
    UnitEntry("Force", "ForceUnit", "pound"),
    UnitEntry("Force", "ForceUnit", "ton"),
    UnitEntry("Frequency", "FrequencyUnit", "gigahertz"),
    UnitEntry("Frequency", "FrequencyUnit", "hertz"),
    UnitEntry("Frequency", "FrequencyUnit", "kilohertz"),
    UnitEntry("Frequency", "FrequencyUnit", "megahertz"),
    UnitEntry("Humidity", "DensityUnit", "gramPerCubicMetre"),
    UnitEntry("Humidity", "DensityUnit", "kilogramPerCubicMetre"),
    UnitEntry("Illuminance", "IlluminanceUnit", "footcandle"),
    UnitEntry("Illuminance", "IlluminanceUnit", "lux"),
    UnitEntry("Inductance", "InductanceUnit", "henry"),
    UnitEntry("Inductance", "InductanceUnit", "microhenry"),
    UnitEntry("Inductance", "InductanceUnit", "millihenry"),
    UnitEntry("Latitude", "AngleUnit", "degreeOfArc"),
    UnitEntry("Latitude", "AngleUnit", "minuteOfArc"),
    UnitEntry("Latitude", "AngleUnit", "radian"),
    UnitEntry("Latitude", "AngleUnit", "secondOfArc"),
    UnitEntry("Latitude", "AngleUnit", "turn"),
    UnitEntry("Length", "LengthUnit", "astronomicalUnit"),
    UnitEntry("Length", "LengthUnit", "centimetre"),
    UnitEntry("Length", "LengthUnit", "foot"),
    UnitEntry("Length", "LengthUnit", "inch"),
    UnitEntry("Length", "LengthUnit", "kilometre"),
    UnitEntry("Length", "LengthUnit", "metre"),
    UnitEntry("Length", "LengthUnit", "micrometre"),
    UnitEntry("Length", "LengthUnit", "mile"),
    UnitEntry("Length", "LengthUnit", "millimetre"),
    UnitEntry("Length", "LengthUnit", "nanometre"),
    UnitEntry("Length", "LengthUnit", "nauticalMile"),
    UnitEntry("Longitude", "AngleUnit", "degreeOfArc"),
    UnitEntry("Longitude", "AngleUnit", "minuteOfArc"),
    UnitEntry("Longitude", "AngleUnit", "radian"),
    UnitEntry("Longitude", "AngleUnit", "secondOfArc"),
    UnitEntry("Longitude", "AngleUnit", "turn"),
    UnitEntry("Luminance", "LuminanceUnit", "candelaPerSquareMetre"),
    UnitEntry("Luminosity", "PowerUnit", "gigawatt"),
    UnitEntry("Luminosity", "PowerUnit", "horsepower"),
    UnitEntry("Luminosity", "PowerUnit", "kilowatt"),
    UnitEntry("Luminosity", "PowerUnit", "kilowattHourPerYear"),
    UnitEntry("Luminosity", "PowerUnit", "megawatt"),
    UnitEntry("Luminosity", "PowerUnit", "microwatt"),
    UnitEntry("Luminosity", "PowerUnit", "milliwatt"),
    UnitEntry("Luminosity", "PowerUnit", "watt"),
    UnitEntry("LuminousFlux", "LuminousFluxUnit", "lumen"),
    UnitEntry("LuminousIntensity", "LuminousIntensityUnit", "candela"),
    UnitEntry("MagneticFlux", "MagneticFluxUnit", "maxwell"),
    UnitEntry("MagneticFlux", "MagneticFluxUnit", "weber"),
    UnitEntry("MagneticInduction", "MagneticInductionUnit", "tesla"),
    UnitEntry("Mass", "MassUnit", "gram"),
    UnitEntry("Mass", "MassUnit", "kilogram"),
    UnitEntry("Mass", "MassUnit", "microgram"),
    UnitEntry("Mass", "MassUnit", "milligram"),
    UnitEntry("Mass", "MassUnit", "slug"),
    UnitEntry("Mass", "MassUnit", "tonne"),
    UnitEntry("MassFlowRate", "MassFlowRateUnit", "gramPerHour"),
    UnitEntry("MassFlowRate", "MassFlowRateUnit", "gramPerSecond"),
    UnitEntry("MassFlowRate", "MassFlowRateUnit", "kilogramPerHour"),
    UnitEntry("MassFlowRate", "MassFlowRateUnit", "kilogramPerSecond"),
    UnitEntry("Power", "PowerUnit", "gigawatt"),
    UnitEntry("Power", "PowerUnit", "horsepower"),
    UnitEntry("Power", "PowerUnit", "kilowatt"),
    UnitEntry("Power", "PowerUnit", "kilowattHour"),
    UnitEntry("Power", "PowerUnit", "megawatt"),
    UnitEntry("Power", "PowerUnit", "microwatt"),
    UnitEntry("Power", "PowerUnit", "milliwatt"),
    UnitEntry("Power", "PowerUnit", "watt"),
    UnitEntry("Pressure", "PressureUnit", "bar"),
    UnitEntry("Pressure", "PressureUnit", "inchesOfMercury"),
    UnitEntry("Pressure", "PressureUnit", "inchesOfWater"),
    UnitEntry("Pressure", "PressureUnit", "kilopascal"),
    UnitEntry("Pressure", "PressureUnit", "millibar"),
    UnitEntry("Pressure", "PressureUnit", "millimetresOfMercury"),
    UnitEntry("Pressure", "PressureUnit", "pascal"),
    UnitEntry("Pressure", "PressureUnit", "poundPerSquareInch"),
    UnitEntry("RelativeHumidity", "Unitless", "percent"),
    UnitEntry("RelativeHumidity", "Unitless", "unity"),
    UnitEntry("Resistance", "ResistanceUnit", "kiloohm"),
    UnitEntry("Resistance", "ResistanceUnit", "megaohm"),
    UnitEntry("Resistance", "ResistanceUnit", "milliohm"),
    UnitEntry("Resistance", "ResistanceUnit", "ohm"),
    UnitEntry("SoundPressure", "SoundPressureUnit", "bel"),
    UnitEntry("SoundPressure", "SoundPressureUnit", "decibel"),
    UnitEntry("Temperature", "TemperatureUnit", "degreeCelsius"),
    UnitEntry("Temperature", "TemperatureUnit", "degreeFahrenheit"),
    UnitEntry("Temperature", "TemperatureUnit", "kelvin"),
    UnitEntry("Thrust", "ForceUnit", "newton"),
    UnitEntry("Thrust", "ForceUnit", "ounce"),
    UnitEntry("Thrust", "ForceUnit", "pound"),
    UnitEntry("Thrust", "ForceUnit", "ton"),
    UnitEntry("TimeSpan", "TimeUnit", "day"),
    UnitEntry("TimeSpan", "TimeUnit", "hour"),
    UnitEntry("TimeSpan", "TimeUnit", "microsecond"),
    UnitEntry("TimeSpan", "TimeUnit", "millisecond"),
    UnitEntry("TimeSpan", "TimeUnit", "minute"),
    UnitEntry("TimeSpan", "TimeUnit", "nanosecond"),
    UnitEntry("TimeSpan", "TimeUnit", "second"),
    UnitEntry("TimeSpan", "TimeUnit", "year"),
    UnitEntry("Torque", "TorqueUnit", "newtonMetre"),
    UnitEntry("Velocity", "VelocityUnit", "centimetrePerSecond"),
    UnitEntry("Velocity", "VelocityUnit", "kilometrePerHour"),
    UnitEntry("Velocity", "VelocityUnit", "kilometrePerSecond"),
    UnitEntry("Velocity", "VelocityUnit", "knot"),
    UnitEntry("Velocity", "VelocityUnit", "metrePerHour"),
    UnitEntry("Velocity", "VelocityUnit", "metrePerSecond"),
    UnitEntry("Velocity", "VelocityUnit", "milePerHour"),
    UnitEntry("Velocity", "VelocityUnit", "milePerSecond"),
    UnitEntry("Voltage", "VoltageUnit", "kilovolt"),
    UnitEntry("Voltage", "VoltageUnit", "megavolt"),
    UnitEntry("Voltage", "VoltageUnit", "microvolt"),
    UnitEntry("Voltage", "VoltageUnit", "millivolt"),
    UnitEntry("Voltage", "VoltageUnit", "volt"),
    UnitEntry("Volume", "VolumeUnit", "cubicCentimetre"),
    UnitEntry("Volume", "VolumeUnit", "cubicFoot"),
    UnitEntry("Volume", "VolumeUnit", "cubicInch"),
    UnitEntry("Volume", "VolumeUnit", "cubicMetre"),
    UnitEntry("Volume", "VolumeUnit", "fluidOunce"),
    UnitEntry("Volume", "VolumeUnit", "gallon"),
    UnitEntry("Volume", "VolumeUnit", "litre"),
    UnitEntry("Volume", "VolumeUnit", "millilitre"),
    UnitEntry("VolumeFlowRate", "VolumeFlowRateUnit", "litrePerHour"),
    UnitEntry("VolumeFlowRate", "VolumeFlowRateUnit", "litrePerSecond"),
    UnitEntry("VolumeFlowRate", "VolumeFlowRateUnit", "millilitrePerHour"),
    UnitEntry("VolumeFlowRate", "VolumeFlowRateUnit", "millilitrePerSecond"),
]

UNIT_TYPE_BY_SEMANTIC_TYPE = {entry.semantic_type: entry.unit_type for entry in _UNIT_ENTRIES}
# This is a bit dangerous, as unit names are not unique across types
# It is only used as a backup in the parsing of units.
ENTRY_BY_UNIT = {entry.unit: entry for entry in _UNIT_ENTRIES}



import logging
import os
from pathlib import Path
from typing import Any

from cognite.client import CogniteClient
from fastapi import FastAPI

from cognite import neat
from cognite.neat.config import Config, configure_logging, copy_examples_to_directory, create_data_dir_structure
from cognite.neat.constants import PACKAGE_DIRECTORY
from cognite.neat.utils.cdf import ServiceCogniteClient
from cognite.neat.utils.utils import get_cognite_client_from_config, get_cognite_client_from_token
from cognite.neat.workflows.cdf_store import CdfStore
from cognite.neat.workflows.manager import WorkflowManager
from cognite.neat.workflows.triggers import TriggerManager

UI_PATH = PACKAGE_DIRECTORY / "app" / "ui" / "neat-app" / "build"


class NeatApp:
    def __init__(self, config: Config, cdf_client: CogniteClient | None = None):
        self.config = config
        self.cdf_client: CogniteClient | None = None
        self.cdf_store: CdfStore | None = None
        self.fast_api_app: FastAPI | None = None
        self.cdf_client = cdf_client

    def set_http_server(self, fast_api_app: FastAPI):
        """Set the http server to be used by the triggers manager"""
        self.fast_api_app = fast_api_app

    def start(self, config: Config | None = None):
        logging.info("Starting NeatApp")
        if config:
            self.config = config
        logging.info("Initializing global objects")
        if not self.cdf_client:
            if isinstance(self.config.cdf_client, ServiceCogniteClient):
                if self.config.cdf_client.client_id:
                    self.cdf_client = get_cognite_client_from_config(self.config.cdf_client)
                else:
                    # If no client_id is provided, we assume that the token is provided instead of secret.
                    self.cdf_client = get_cognite_client_from_token(self.config.cdf_client)
            else:
                raise ValueError("Only ServiceCogniteClient is supported at the moment.")
        self.cdf_store = CdfStore(
            self.cdf_client,
            self.config.cdf_default_dataset_id,
            workflows_storage_path=self.config.workflows_store_path,
            rules_storage_path=self.config.rules_store_path,
        )

        # Automatically downloading workflows from CDF if enabled in config
        if self.config.workflow_downloader_filter:
            self.cdf_store.load_workflows_from_cfg_by_filter(self.config.workflow_downloader_filter)

        self.workflow_manager = WorkflowManager(
            client=self.cdf_client,
            config=self.config,
        )
        self.workflow_manager.load_workflows_from_storage()
        self.triggers_manager = TriggerManager(workflow_manager=self.workflow_manager)
        self.triggers_manager.start_time_schedulers()
        logging.info("NeatApp started")

    def stop(self):
        logging.info("Stopping NeatApp")
        if self.triggers_manager:
            self.triggers_manager.stop_scheduler_main_loop()
        self.cdf_client = None
        self.cdf_store = None
        self.workflow_manager = None
        self.triggers_manager = None
        logging.info("NeatApp stopped")


def create_neat_app() -> NeatApp:
    logger = logging.getLogger(__name__)  # temporary logger before config is loaded

    if os.environ.get("NEAT_CDF_PROJECT"):
        logger.info("ENV NEAT_CDF_PROJECT is set, loading config from env.")
        config = Config.from_env()
    elif (config_path := Path(os.environ.get("NEAT_CONFIG_PATH", "config.yaml"))).exists():
        logger.info(f"Loading config from {config_path.name}.")
        config = Config.from_yaml(config_path)
    else:
        logger.error(f"Config file {config_path.name} not found.Exiting.")
        config = Config()
        config.to_yaml(config_path)

    if config.load_examples:
        copy_examples_to_directory(config)
    else:
        create_data_dir_structure(config)

    configure_logging(config.log_level, config.log_format)
    logging.info(f" Starting NEAT version {neat.__version__}")
    logging.debug(f" Config: {config.model_dump(exclude={'cdf_client': {'client_secret': ...}})}")

    return NeatApp(config)


NEAT_APP = create_neat_app()
CACHE_STORE: dict[str, Any] = {}
